"""
Unit tests for JuliaExplicitComp
"""
import os
import time
import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal

# from juliacall import Main as jl
import juliacall; jl = juliacall.newmodule("OpenMDAOJuliaExplicitCompTest")

from omjlcomps import JuliaExplicitComp

d = os.path.dirname(os.path.abspath(__file__))
jl.include(os.path.join(d, "test_ecomp.jl"))


class TestSimpleJuliaExplicitComp(unittest.TestCase):

    def setUp(self):
        p = self.p = om.Problem()
        ecomp = jl.ECompTest.ECompSimple()
        comp = JuliaExplicitComp(jlcomp=ecomp)
        p.model.add_subsystem("ecomp", comp, promotes_inputs=["x"], promotes_outputs=["y"])
        p.setup(force_alloc_complex=True)
        p.set_val("x", 3.0)
        p.run_model()

    def test_results(self):
        p = self.p
        expected = 2*p.get_val("x")[0]**2 + 1
        actual = p.get_val("y")[0]
        np.testing.assert_almost_equal(actual, expected)

    def test_partials(self):
        p = self.p
        np.set_printoptions(linewidth=1024)
        cpd = self.p.check_partials(compact_print=True, out_stream=None, method='cs')

        # Check that the partials the user provided are correct.
        ecomp_partials = cpd["ecomp"]
        np.testing.assert_almost_equal(actual=ecomp_partials["y", "x"]['J_fwd'], desired=[[4*p.get_val("x")[0]]], decimal=12)

        # Check that partials approximated by the complex-step method match the user-provided partials.
        for comp in cpd:
            for (var, wrt) in cpd[comp]:
                np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_fwd'],
                                               desired=cpd[comp][var, wrt]['J_fd'],
                                               decimal=12)

class TestSimpleJuliaExplicitCSComp(unittest.TestCase):

    def setUp(self):
        p = self.p = om.Problem()
        ecomp = jl.ECompTest.ECompSimpleCS()
        comp = JuliaExplicitComp(jlcomp=ecomp)
        p.model.add_subsystem("ecomp", comp, promotes_inputs=["x"], promotes_outputs=["y"])
        p.setup(force_alloc_complex=True)
        p.set_val("x", 3.0)
        p.run_model()

    def test_results(self):
        p = self.p
        expected = 2*p.get_val("x")[0]**2 + 1
        actual = p.get_val("y")[0]
        np.testing.assert_almost_equal(actual, expected)

    def test_partials(self):
        p = self.p
        np.set_printoptions(linewidth=1024)
        cpd = self.p.check_partials(compact_print=True, out_stream=None, method='fd')

        # Check that the partials calculated by OpenMDAO's complex step match the exact.
        ecomp_partials = cpd["ecomp"]
        np.testing.assert_almost_equal(actual=ecomp_partials["y", "x"]['J_fwd'], desired=[[4*p.get_val("x")[0]]], decimal=12)

        # Check that partials approximated by the complex-step method match the finite difference check.
        for comp in cpd:
            for (var, wrt) in cpd[comp]:
                np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_fwd'],
                                               desired=cpd[comp][var, wrt]['J_fd'],
                                               decimal=5)

class TestJuliaExplicitCompWithOption(unittest.TestCase):

    def setUp(self):
        p = self.p = om.Problem()
        a = self.a = 0.5
        ecomp = jl.ECompTest.ECompWithOption(a)
        comp = JuliaExplicitComp(jlcomp=ecomp)
        p.model.add_subsystem("ecomp", comp, promotes_inputs=["x"], promotes_outputs=["y"])
        p.setup(force_alloc_complex=True)
        p.set_val("x", 3.0)
        p.run_model()

    def test_results(self):
        p = self.p
        expected = 2*self.a*p.get_val("x")[0]**2 + 1
        actual = p.get_val("y")[0]
        np.testing.assert_almost_equal(actual, expected)

    def test_partials(self):
        p = self.p
        np.set_printoptions(linewidth=1024)
        cpd = self.p.check_partials(compact_print=True, out_stream=None, method='cs')

        # Check that the partials the user provided are correct.
        ecomp_partials = cpd["ecomp"]
        np.testing.assert_almost_equal(actual=ecomp_partials["y", "x"]['J_fwd'], desired=[[4*self.a*p.get_val("x")[0]]], decimal=12)

        # Check that partials approximated by the complex-step method match the user-provided partials.
        for comp in cpd:
            for (var, wrt) in cpd[comp]:
                np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_fwd'],
                                               desired=cpd[comp][var, wrt]['J_fd'],
                                               decimal=12)

class TestJuliaExplicitCompWithOptionAndTags(unittest.TestCase):

    def setUp(self):
        p = self.p = om.Problem()
        a = self.a = 0.5
        ecomp = jl.ECompTest.ECompWithOptionAndTags(a, xtag=juliacall.convert(jl.Vector, ["xtag1", "xtag2"]), ytag=juliacall.convert(jl.Vector, ["ytag"]))
        comp = JuliaExplicitComp(jlcomp=ecomp)
        p.model.add_subsystem("ecomp", comp, promotes_inputs=["x"], promotes_outputs=["y"])
        p.setup(force_alloc_complex=True)
        p.set_val("x", 3.0)
        p.run_model()

    def test_results1(self):
        p = self.p
        expected = 2*self.a*p.get_val("x")[0]**2 + 1
        actual = p.get_val("y")[0]
        np.testing.assert_almost_equal(actual, expected)

    def test_partials1(self):
        p = self.p
        np.set_printoptions(linewidth=1024)
        cpd = self.p.check_partials(compact_print=True, out_stream=None, method='cs')

        # Check that the partials the user provided are correct.
        ecomp_partials = cpd["ecomp"]
        np.testing.assert_almost_equal(actual=ecomp_partials["y", "x"]['J_fwd'], desired=[[4*self.a*p.get_val("x")[0]]], decimal=12)

        # Check that partials approximated by the complex-step method match the user-provided partials.
        for comp in cpd:
            for (var, wrt) in cpd[comp]:
                np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_fwd'],
                                               desired=cpd[comp][var, wrt]['J_fd'],
                                               decimal=12)

class TestJuliaExplicitCompWithLargeOption(unittest.TestCase):

    def setUp(self):
        n_small = self.n_small = 10
        p_small = self.p_small = om.Problem()
        ecomp_small = jl.ECompTest.ECompWithLargeOption(n_small)
        comp_small = JuliaExplicitComp(jlcomp=ecomp_small)
        p_small.model.add_subsystem("ecomp", comp_small, promotes_inputs=["x"], promotes_outputs=["y"])
        p_small.setup(force_alloc_complex=True)
        p_small.set_val("x", 3.0)
        p_small.run_model()

        n_big = self.n_big = 1_000_000_000
        p_big = self.p_big = om.Problem()
        ecomp_big = jl.ECompTest.ECompWithLargeOption(n_big)
        comp_big = JuliaExplicitComp(jlcomp=ecomp_big)
        p_big.model.add_subsystem("ecomp", comp_big, promotes_inputs=["x"], promotes_outputs=["y"])
        p_big.setup(force_alloc_complex=True)
        p_big.set_val("x", 3.0)
        p_big.run_model()

    def test_results(self):
        for p in [self.p_small, self.p_big]:
            expected = 2*3*p.get_val("x")[0]**2 + 1
            actual = p.get_val("y")[0]
            np.testing.assert_almost_equal(actual, expected)

    def test_partials(self):
        for p in [self.p_small, self.p_big]:
            np.set_printoptions(linewidth=1024)
            cpd = p.check_partials(compact_print=True, out_stream=None, method='cs')

            # Check that the partials the user provided are correct.
            ecomp_partials = cpd["ecomp"]
            np.testing.assert_almost_equal(actual=ecomp_partials["y", "x"]['J_fwd'], desired=[[4*3*p.get_val("x")[0]]], decimal=12)

            # Check that partials approximated by the complex-step method match the user-provided partials.
            for comp in cpd:
                for (var, wrt) in cpd[comp]:
                    np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_fwd'],
                                                   desired=cpd[comp][var, wrt]['J_fd'],
                                                   decimal=12)

    def test_timings(self):
        time_avg = []
        n_samples = 1000
        for p in [self.p_small, self.p_big]:
            # Just to make sure the Julia JIT is all warmed up.
            p.run_model()
            tavg = 0.0
            for sample in range(n_samples):
                tstart = time.time()
                p.run_model()
                telapsed = time.time() - tstart
                tavg += telapsed/n_samples

            time_avg.append(tavg)

        # Compare the average timings.
        np.testing.assert_almost_equal(time_avg[1]/time_avg[0], 1.0, decimal=1)


class TestJuliaMatrixFreeExplicitComp(unittest.TestCase):

    def setUp(self):
        p = self.p = om.Problem()
        nrows, ncols = 2, 3
        ecomp = jl.ECompTest.ECompMatrixFree(nrows, ncols)
        comp = JuliaExplicitComp(jlcomp=ecomp)
        p.model.add_subsystem("ecomp", comp, promotes_inputs=["x1", "x2"], promotes_outputs=["y1", "y2"])
        p.setup(force_alloc_complex=True)
        p.set_val("x1", np.arange(nrows*ncols).reshape(nrows, ncols)+0.5)
        p.set_val("x2", np.arange(nrows*ncols).reshape(nrows, ncols)+1)
        p.run_model()

    def test_results(self):
        p = self.p
        expected = 2*p.get_val("x1") + 3*p.get_val("x2")**2
        actual = p.get_val("y1")
        np.testing.assert_almost_equal(actual, expected)

        expected = 4*p.get_val("x1")**3 + 5*p.get_val("x2")**4
        actual = p.get_val("y2")
        np.testing.assert_almost_equal(actual, expected)

    def test_partials(self):
        p = self.p
        np.set_printoptions(linewidth=1024)
        cpd = self.p.check_partials(compact_print=True, out_stream=None, method='cs')

        # Check that partials approximated by the complex-step method match the user-provided partials.
        for comp in cpd:
            for (var, wrt) in cpd[comp]:
                np.testing.assert_allclose(actual=cpd[comp][var, wrt]['J_fwd'], desired=cpd[comp][var, wrt]['J_fd'], rtol=1e-12)
                np.testing.assert_allclose(actual=cpd[comp][var, wrt]['J_rev'], desired=cpd[comp][var, wrt]['J_fd'], rtol=1e-12)

        p.set_val("x1", np.arange(2*3).reshape(2,3)+4)
        p.set_val("x2", np.arange(2*3).reshape(2,3)+5)

        cpd = self.p.check_partials(compact_print=True, out_stream=None, method='cs')

        # Check that partials approximated by the complex-step method match the user-provided partials.
        for comp in cpd:
            for (var, wrt) in cpd[comp]:
                np.testing.assert_allclose(actual=cpd[comp][var, wrt]['J_fwd'], desired=cpd[comp][var, wrt]['J_fd'], rtol=1e-12)
                np.testing.assert_allclose(actual=cpd[comp][var, wrt]['J_rev'], desired=cpd[comp][var, wrt]['J_fd'], rtol=1e-12)


if __name__ == '__main__':
    unittest.main()



"""
Unit tests for JuliaImplicitComp
"""
import os
import time
import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal

# from juliacall import Main as jl
import juliacall; jl = juliacall.newmodule("OpenMDAOJuliaImplicitCompTest")

from omjlcomps import JuliaImplicitComp

d = os.path.dirname(os.path.abspath(__file__))
jl.include(os.path.join(d, "test_icomp.jl"))


class TestSimpleJuliaImplicitComp(unittest.TestCase):

    def setUp(self):
        p = self.p = om.Problem()
        n = self.n = 10
        a = self.a = 3.0
        icomp = jl.ICompTest.SimpleImplicit(n, a)
        comp = JuliaImplicitComp(jlcomp=icomp)
        comp.linear_solver = om.DirectSolver(assemble_jac=True)
        comp.nonlinear_solver = om.NewtonSolver(solve_subsystems=True, iprint=2, err_on_non_converge=True)
        p.model.add_subsystem("icomp", comp, promotes_inputs=["x", "y"], promotes_outputs=["z1", "z2"])
        p.setup(force_alloc_complex=True)
        p.set_val("x", 3.0)
        p.set_val("y", 4.0)
        p.run_model()

    def test_results(self):
        p = self.p
        a = self.a
        expected = a*p.get_val("x")**2 + p.get_val("y")**2
        actual = p.get_val("z1")
        np.testing.assert_almost_equal(actual, expected)
        expected = a*p.get_val("x") + p.get_val("y")
        actual = p.get_val("z2")
        np.testing.assert_almost_equal(actual, expected)

    def test_partials(self):
        p = self.p
        np.set_printoptions(linewidth=1024)
        cpd = self.p.check_partials(compact_print=True, out_stream=None, method='cs')

        # Check that the partials the user provided are correct.
        icomp_partials = cpd["icomp"]

        actual = icomp_partials["z1", "x"]['J_fwd']
        expected = np.zeros((self.n, self.n))
        expected[range(self.n), range(self.n)] = 2*self.a*p.get_val("x")
        np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

        actual = icomp_partials["z1", "y"]['J_fwd']
        expected = np.zeros((self.n, self.n))
        expected[range(self.n), range(self.n)] = 2*p.get_val("y")
        np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

        actual = icomp_partials["z1", "z1"]['J_fwd']
        expected = np.zeros((self.n, self.n))
        expected[range(self.n), range(self.n)] = -1.0
        np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

        actual = icomp_partials["z2", "x"]['J_fwd']
        expected = np.zeros((self.n, self.n))
        expected[range(self.n), range(self.n)] = self.a
        np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

        actual = icomp_partials["z2", "y"]['J_fwd']
        expected = np.zeros((self.n, self.n))
        expected[range(self.n), range(self.n)] = 1.0
        np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

        actual = icomp_partials["z2", "z2"]['J_fwd']
        expected = np.zeros((self.n, self.n))
        expected[range(self.n), range(self.n)] = -1.0
        np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

        # Check that partials approximated by the complex-step method match the user-provided partials.
        for comp in cpd:
            for (var, wrt) in cpd[comp]:
                np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_fwd'],
                                               desired=cpd[comp][var, wrt]['J_fd'],
                                               decimal=12)


class TestSolveNonlinearJuliaImplicitComp(unittest.TestCase):

    def setUp(self):
        p = self.p = om.Problem()
        n = self.n = 10
        a = self.a = 3.0
        icomp = jl.ICompTest.SolveNonlinearImplicit(n, a)
        comp = JuliaImplicitComp(jlcomp=icomp)
        comp.linear_solver = om.DirectSolver(assemble_jac=True)
        p.model.add_subsystem("icomp", comp, promotes_inputs=["x", "y"], promotes_outputs=["z1", "z2"])
        p.setup(force_alloc_complex=True)
        p.set_val("x", 3.0)
        p.set_val("y", 4.0)
        p.run_model()

    def test_results(self):
        p = self.p
        a = self.a
        expected = a*p.get_val("x")**2 + p.get_val("y")**2
        actual = p.get_val("z1")
        np.testing.assert_almost_equal(actual, expected)
        expected = a*p.get_val("x") + p.get_val("y")
        actual = p.get_val("z2")
        np.testing.assert_almost_equal(actual, expected)

    def test_partials(self):
        p = self.p
        np.set_printoptions(linewidth=1024)
        cpd = self.p.check_partials(compact_print=True, out_stream=None, method='cs')

        # Check that the partials the user provided are correct.
        icomp_partials = cpd["icomp"]

        actual = icomp_partials["z1", "x"]['J_fwd']
        expected = np.zeros((self.n, self.n))
        expected[range(self.n), range(self.n)] = 2*self.a*p.get_val("x")
        np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

        actual = icomp_partials["z1", "y"]['J_fwd']
        expected = np.zeros((self.n, self.n))
        expected[range(self.n), range(self.n)] = 2*p.get_val("y")
        np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

        actual = icomp_partials["z1", "z1"]['J_fwd']
        expected = np.zeros((self.n, self.n))
        expected[range(self.n), range(self.n)] = -1.0
        np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

        actual = icomp_partials["z2", "x"]['J_fwd']
        expected = np.zeros((self.n, self.n))
        expected[range(self.n), range(self.n)] = self.a
        np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

        actual = icomp_partials["z2", "y"]['J_fwd']
        expected = np.zeros((self.n, self.n))
        expected[range(self.n), range(self.n)] = 1.0
        np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

        actual = icomp_partials["z2", "z2"]['J_fwd']
        expected = np.zeros((self.n, self.n))
        expected[range(self.n), range(self.n)] = -1.0
        np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

        # Check that partials approximated by the complex-step method match the user-provided partials.
        for comp in cpd:
            for (var, wrt) in cpd[comp]:
                np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_fwd'],
                                               desired=cpd[comp][var, wrt]['J_fd'],
                                               decimal=12)


class TestMatrixFreeImplicitComp(unittest.TestCase):

    def setUp(self):
        p = self.p = om.Problem()
        n = self.n = 10
        a = self.a = 3.0
        icomp = jl.ICompTest.MatrixFreeImplicit(n, a)
        comp = JuliaImplicitComp(jlcomp=icomp)
        comp.linear_solver = om.DirectSolver(assemble_jac=False)
        comp.nonlinear_solver = om.NewtonSolver(solve_subsystems=True, iprint=2, err_on_non_converge=True)
        p.model.add_subsystem("icomp", comp, promotes_inputs=["x", "y"], promotes_outputs=["z1", "z2"])
        p.setup(force_alloc_complex=True)
        p.set_val("x", np.arange(n)+0.5)
        p.set_val("y", np.arange(n)+2)
        p.run_model()

    def test_results(self):
        p = self.p
        a = self.a
        expected = a*p.get_val("x")**2 + p.get_val("y")**2
        actual = p.get_val("z1")
        np.testing.assert_almost_equal(actual, expected)
        expected = a*p.get_val("x") + p.get_val("y")
        actual = p.get_val("z2")
        np.testing.assert_almost_equal(actual, expected)

    def test_partials(self):
        p = self.p
        np.set_printoptions(linewidth=1024)
        cpd = self.p.check_partials(compact_print=True, out_stream=None, method='cs')

        # Check that the partials the user provided are correct.
        icomp_partials = cpd["icomp"]

        actual = icomp_partials["z1", "x"]['J_fwd']
        expected = np.zeros((self.n, self.n))
        expected[range(self.n), range(self.n)] = 2*self.a*p.get_val("x")
        np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

        actual = icomp_partials["z1", "y"]['J_fwd']
        expected = np.zeros((self.n, self.n))
        expected[range(self.n), range(self.n)] = 2*p.get_val("y")
        np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

        actual = icomp_partials["z1", "z1"]['J_fwd']
        expected = np.zeros((self.n, self.n))
        expected[range(self.n), range(self.n)] = -1.0
        np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

        actual = icomp_partials["z2", "x"]['J_fwd']
        expected = np.zeros((self.n, self.n))
        expected[range(self.n), range(self.n)] = self.a
        np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

        actual = icomp_partials["z2", "y"]['J_fwd']
        expected = np.zeros((self.n, self.n))
        expected[range(self.n), range(self.n)] = 1.0
        np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

        actual = icomp_partials["z2", "z2"]['J_fwd']
        expected = np.zeros((self.n, self.n))
        expected[range(self.n), range(self.n)] = -1.0
        np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

        # Check that partials approximated by the complex-step method match the user-provided partials.
        for comp in cpd:
            for (var, wrt) in cpd[comp]:
                np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_fwd'],
                                               desired=cpd[comp][var, wrt]['J_fd'],
                                               decimal=12)
                np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_rev'],
                                               desired=cpd[comp][var, wrt]['J_fd'],
                                               decimal=12)


class TestSolveLinearImplicitComp(unittest.TestCase):

    def setUp(self):
        n = self.n = 10
        a = self.a = 3.0

        p = self.p_fwd = om.Problem()
        icomp = jl.ICompTest.SolveLinearImplicit(n, a)
        comp = JuliaImplicitComp(jlcomp=icomp)
        p.model.add_subsystem("icomp", comp, promotes_inputs=["x", "y"], promotes_outputs=["z1", "z2"])

        p.setup(force_alloc_complex=True, mode="fwd")
        p.set_val("x", np.arange(n)+0.5)
        p.set_val("y", np.arange(n)+2)
        p.run_model()

        p = self.p_rev = om.Problem()
        icomp = jl.ICompTest.SolveLinearImplicit(n, a)
        comp = JuliaImplicitComp(jlcomp=icomp)
        p.model.add_subsystem("icomp", comp, promotes_inputs=["x", "y"], promotes_outputs=["z1", "z2"])

        p.setup(force_alloc_complex=True, mode="rev")
        p.set_val("x", np.arange(n)+0.5)
        p.set_val("y", np.arange(n)+2)
        p.run_model()

    def test_results(self):
        for p in [self.p_fwd, self.p_rev]:
            a = self.a
            expected = a*p.get_val("x")**2 + p.get_val("y")**2
            actual = p.get_val("z1")
            np.testing.assert_almost_equal(actual, expected)
            expected = a*p.get_val("x") + p.get_val("y")
            actual = p.get_val("z2")
            np.testing.assert_almost_equal(actual, expected)

    def test_partials(self):
        for p in [self.p_fwd, self.p_rev]:
            np.set_printoptions(linewidth=1024)
            cpd = p.check_partials(compact_print=True, out_stream=None, method='cs')

            # Check that the partials the user provided are correct.
            icomp_partials = cpd["icomp"]

            actual = icomp_partials["z1", "x"]['J_fwd']
            expected = np.zeros((self.n, self.n))
            expected[range(self.n), range(self.n)] = 2*self.a*p.get_val("x")
            np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

            actual = icomp_partials["z1", "y"]['J_fwd']
            expected = np.zeros((self.n, self.n))
            expected[range(self.n), range(self.n)] = 2*p.get_val("y")
            np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

            actual = icomp_partials["z1", "z1"]['J_fwd']
            expected = np.zeros((self.n, self.n))
            expected[range(self.n), range(self.n)] = -1.0
            np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

            actual = icomp_partials["z2", "x"]['J_fwd']
            expected = np.zeros((self.n, self.n))
            expected[range(self.n), range(self.n)] = self.a
            np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

            actual = icomp_partials["z2", "y"]['J_fwd']
            expected = np.zeros((self.n, self.n))
            expected[range(self.n), range(self.n)] = 1.0
            np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

            actual = icomp_partials["z2", "z2"]['J_fwd']
            expected = np.zeros((self.n, self.n))
            expected[range(self.n), range(self.n)] = -1.0
            np.testing.assert_almost_equal(actual=actual, desired=expected, decimal=12)

            # Check that partials approximated by the complex-step method match the user-provided partials.
            for comp in cpd:
                for (var, wrt) in cpd[comp]:
                    np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_fwd'],
                                                   desired=cpd[comp][var, wrt]['J_fd'],
                                                   decimal=12)
                    # np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_rev'],
                    #                                desired=cpd[comp][var, wrt]['J_fd'],
                    #                                decimal=12)

    def test_totals(self):
        for p in [self.p_fwd, self.p_rev]:
            ctd = p.check_totals(of=["z1", "z2"], wrt=["x", "y"], method='cs', compact_print=True, out_stream=None)
            for of_wrt in ctd:
                np.testing.assert_almost_equal(actual=ctd[of_wrt]['J_fwd'],
                                               desired=ctd[of_wrt]['J_fd'],
                                               decimal=12)



class TestGuessNonlinearImplicitComp(unittest.TestCase):

    def setUp(self):
        n = self.n = 1

        # Create a component that should find the left root of R(x) = x**2 - 4*x + 3 = (x - 1)*(x - 3) = 0, aka 1.
        # (x - 1)*(x - 3)
        xguess = 1.5
        xlower = -10.0
        xupper = 1.9
        p_leftroot = self.p_leftroot = om.Problem()
        icomp = jl.ICompTest.GuessNonlinearImplicit(n, xguess, xlower, xupper)
        comp = JuliaImplicitComp(jlcomp=icomp)
        comp.linear_solver = om.DirectSolver(assemble_jac=True)
        comp.nonlinear_solver = om.NewtonSolver(solve_subsystems=True, iprint=2, err_on_non_converge=True)
        comp.nonlinear_solver.linesearch = om.BoundsEnforceLS(bound_enforcement='scalar')
        p_leftroot.model.add_subsystem("icomp", comp, promotes_inputs=["a", "b", "c"], promotes_outputs=["x"])
        p_leftroot.setup(force_alloc_complex=True)
        p_leftroot.set_val("a", 1.0)
        p_leftroot.set_val("b", -4.0)
        p_leftroot.set_val("c", 3.0)
        p_leftroot.run_model()

        # Create a component that should find the right root of R(x) = x**2 - 4*x + 3 = 0, aka 3.
        xguess = 2.5
        xlower = 2.1
        xupper = 10.0
        p_rightroot = self.p_rightroot = om.Problem()
        icomp = jl.ICompTest.GuessNonlinearImplicit(n, xguess, xlower, xupper)
        comp = JuliaImplicitComp(jlcomp=icomp)
        comp.linear_solver = om.DirectSolver(assemble_jac=True)
        comp.nonlinear_solver = om.NewtonSolver(solve_subsystems=True, iprint=2, err_on_non_converge=True)
        comp.nonlinear_solver.linesearch = om.BoundsEnforceLS(bound_enforcement='scalar')
        p_rightroot.model.add_subsystem("icomp", comp, promotes_inputs=["a", "b", "c"], promotes_outputs=["x"])
        p_rightroot.setup(force_alloc_complex=True)
        p_rightroot.set_val("a", 1.0)
        p_rightroot.set_val("b", -4.0)
        p_rightroot.set_val("c", 3.0)
        p_rightroot.run_model()

    def test_results(self):
        expected = 1*np.ones(self.n)
        actual = self.p_leftroot.get_val("x")
        np.testing.assert_almost_equal(actual, expected)

        expected = 3*np.ones(self.n)
        actual = self.p_rightroot.get_val("x")
        np.testing.assert_almost_equal(actual, expected)

    def test_partials(self):
        np.set_printoptions(linewidth=1024)
        for p in [self.p_leftroot, self.p_rightroot]:
            cpd = p.check_partials(compact_print=True, out_stream=None, method='cs')

            # Check that partials approximated by the complex-step method match the user-provided partials.
            for comp in cpd:
                for (var, wrt) in cpd[comp]:
                    np.testing.assert_almost_equal(actual=cpd[comp][var, wrt]['J_fwd'],
                                                   desired=cpd[comp][var, wrt]['J_fd'],
                                                   decimal=12)



if __name__ == '__main__':
    unittest.main()



from types import MethodType

import openmdao.api as om
from openmdao.core.analysis_error import AnalysisError

import juliacall; jl = juliacall.newmodule("OpenMDAOJuliaComps")
from juliacall import JuliaError
# This imports the Julia package OpenMDAOCore:
jl.seval("using OpenMDAOCore: OpenMDAOCore")


def _initialize_common(self):
    self.options.declare('jlcomp')


def _setup_common(self):
    self._jlcomp = self.options['jlcomp']
    input_data, output_data, partials_data = jl.OpenMDAOCore.setup(self._jlcomp)

    for var in input_data:
        if var.tags is not None:
            tags = list(var.tags)
        else:
            tags = None
        self.add_input(var.name, shape=var.shape, val=var.val,
                       units=var.units, tags=tags)

    for var in output_data:
        if var.tags is not None:
            tags = list(var.tags)
        else:
            tags = None
        self.add_output(var.name, shape=var.shape, val=var.val,
                        units=var.units, lower=var.lower, upper=var.upper, tags=tags)

    for data in partials_data:
        self.declare_partials(data.of, data.wrt,
                              rows=data.rows, cols=data.cols,
                              val=data.val, method=data.method)


class JuliaExplicitComp(om.ExplicitComponent):

    def initialize(self):
        _initialize_common(self)

    def setup(self):
        _setup_common(self)

        if jl.OpenMDAOCore.has_compute_partials(self._jlcomp):
            def compute_partials(self, inputs, partials):
                inputs_dict = juliacall.convert(jl.Dict, dict(inputs))

                partials_dict = {}
                for of_wrt in self._declared_partials:
                    partials_dict[of_wrt] = partials[of_wrt]
                partials_dict = juliacall.convert(jl.Dict, partials_dict)

                try:
                    jl.OpenMDAOCore.compute_partials_b(self._jlcomp, inputs_dict, partials_dict)
                except JuliaError as e:
                    if jl.isa(e.exception, jl.DomainError):
                        raise AnalysisError(f"caught Julia DomainError in {self}.compute_partials:\n{e}")
                    else:
                        raise e from None

            # https://www.ianlewis.org/en/dynamically-adding-method-classes-or-class-instanc
            self.compute_partials = MethodType(compute_partials, self)
            # Hmm...
            self._has_compute_partials = True

        if jl.OpenMDAOCore.has_compute_jacvec_product(self._jlcomp):
            def compute_jacvec_product(self, inputs, d_inputs, d_outputs, mode):
                inputs_dict = juliacall.convert(jl.Dict, dict(inputs))
                d_inputs_dict = juliacall.convert(jl.Dict, dict(d_inputs))
                d_outputs_dict = juliacall.convert(jl.Dict, dict(d_outputs))

                try:
                    jl.OpenMDAOCore.compute_jacvec_product_b(self._jlcomp, inputs_dict, d_inputs_dict, d_outputs_dict, mode)
                except JuliaError as e:
                    if jl.isa(e.exception, jl.DomainError):
                        raise AnalysisError(f"caught Julia DomainError in {self}.compute_jacvec_product:\n{e}")
                    else:
                        raise e from None

            self.compute_jacvec_product = MethodType(compute_jacvec_product, self)

    def compute(self, inputs, outputs):
        inputs_dict = juliacall.convert(jl.Dict, dict(inputs))
        outputs_dict = juliacall.convert(jl.Dict, dict(outputs))

        try:
            jl.OpenMDAOCore.compute_b(self._jlcomp, inputs_dict, outputs_dict)
        except JuliaError as e:
            if jl.isa(e.exception, jl.DomainError):
                raise AnalysisError(f"caught Julia DomainError in {self}.compute:\n{e}")
            else:
                raise e from None


class JuliaImplicitComp(om.ImplicitComponent):

    def initialize(self):
        _initialize_common(self)

    def setup(self):
        _setup_common(self)

        if jl.OpenMDAOCore.has_apply_nonlinear(self._jlcomp):
            def apply_nonlinear(self, inputs, outputs, residuals):
                inputs_dict = juliacall.convert(jl.Dict, dict(inputs))
                outputs_dict = juliacall.convert(jl.Dict, dict(outputs))
                residuals_dict = juliacall.convert(jl.Dict, dict(residuals))

                try:
                    jl.OpenMDAOCore.apply_nonlinear_b(self._jlcomp, inputs_dict, outputs_dict, residuals_dict)
                except JuliaError as e:
                    if jl.isa(e.exception, jl.DomainError):
                        raise AnalysisError(f"caught Julia DomainError in {self}.apply_nonlinear:\n{e}")
                    else:
                        raise e from None

            self.apply_nonlinear = MethodType(apply_nonlinear, self)

        if jl.OpenMDAOCore.has_solve_nonlinear(self._jlcomp):
            def solve_nonlinear(self, inputs, outputs):
                inputs_dict = juliacall.convert(jl.Dict, dict(inputs))
                outputs_dict = juliacall.convert(jl.Dict, dict(outputs))

                try:
                    jl.OpenMDAOCore.solve_nonlinear_b(self._jlcomp, inputs_dict, outputs_dict)
                except JuliaError as e:
                    if jl.isa(e.exception, jl.DomainError):
                        raise AnalysisError(f"caught Julia DomainError in {self}.solve_nonlinear:\n{e}")
                    else:
                        raise e from None

            self.solve_nonlinear = MethodType(solve_nonlinear, self)

        if jl.OpenMDAOCore.has_linearize(self._jlcomp):
            def linearize(self, inputs, outputs, partials):
                inputs_dict = juliacall.convert(jl.Dict, dict(inputs))
                outputs_dict = juliacall.convert(jl.Dict, dict(outputs))

                partials_dict = {}
                for of_wrt in self._declared_partials:
                    partials_dict[of_wrt] = partials[of_wrt]
                partials_dict = juliacall.convert(jl.Dict, partials_dict)

                try:
                    jl.OpenMDAOCore.linearize_b(self._jlcomp, inputs_dict, outputs_dict, partials_dict)
                except JuliaError as e:
                    if jl.isa(e.exception, jl.DomainError):
                        raise AnalysisError(f"caught Julia DomainError in {self}.linearize:\n{e}")
                    else:
                        raise e from None

            self.linearize = MethodType(linearize, self)

        if jl.OpenMDAOCore.has_apply_linear(self._jlcomp):
            def apply_linear(self, inputs, outputs, d_inputs, d_outputs, d_residuals, mode):
                inputs_dict = juliacall.convert(jl.Dict, dict(inputs))
                outputs_dict = juliacall.convert(jl.Dict, dict(outputs))
                d_inputs_dict = juliacall.convert(jl.Dict, dict(d_inputs))
                d_outputs_dict = juliacall.convert(jl.Dict, dict(d_outputs))
                d_residuals_dict = juliacall.convert(jl.Dict, dict(d_residuals))

                try:
                    jl.OpenMDAOCore.apply_linear_b(self._jlcomp, inputs_dict, outputs_dict,
                            d_inputs_dict, d_outputs_dict, d_residuals_dict, mode)
                except JuliaError as e:
                    if jl.isa(e.exception, jl.DomainError):
                        raise AnalysisError(f"caught Julia DomainError in {self}.apply_linear:\n{e}")
                    else:
                        raise e from None

            self.apply_linear = MethodType(apply_linear, self)

        if jl.OpenMDAOCore.has_solve_linear(self._jlcomp):
            def solve_linear(self, d_outputs, d_residuals, mode):
                d_outputs_dict = juliacall.convert(jl.Dict, dict(d_outputs))
                d_residuals_dict = juliacall.convert(jl.Dict, dict(d_residuals))

                try:
                    jl.OpenMDAOCore.solve_linear_b(self._jlcomp, d_outputs_dict, d_residuals_dict, mode)
                except JuliaError as e:
                    if jl.isa(e.exception, jl.DomainError):
                        raise AnalysisError(f"caught Julia DomainError in {self}.solve_linear:\n{e}")
                    else:
                        raise e from None

            # Hello Owen. Red Knights vs Greens.
            self.solve_linear = MethodType(solve_linear, self)


    def _configure(self):
        super()._configure()

        if jl.OpenMDAOCore.has_guess_nonlinear(self._jlcomp):
            def guess_nonlinear(self, inputs, outputs, residuals):
                inputs_dict = juliacall.convert(jl.Dict, dict(inputs))
                outputs_dict = juliacall.convert(jl.Dict, dict(outputs))
                residuals_dict = juliacall.convert(jl.Dict, dict(residuals))

                try:
                    jl.OpenMDAOCore.guess_nonlinear_b(self._jlcomp, inputs_dict, outputs_dict, residuals_dict)
                except JuliaError as e:
                    if jl.isa(e.exception, jl.DomainError):
                        raise AnalysisError(f"caught Julia DomainError in {self}.guess_nonlinear:\n{e}")
                    else:
                        raise e from None

            self.guess_nonlinear = MethodType(guess_nonlinear, self)
            # Hmm...
            self._has_guess = True




from setuptools import setup

setup_args = {'description': 'Create OpenMDAO Components using the Julia programming language',
   'entry_points': {
       'openmdao_component': [
           'juliaexplicitcomp=omjlcomps:JuliaExplicitComp',
           'juliaimplicitcomp=omjlcomps:JuliaImplicitComp'
       ]
    },
   'install_requires': ['openmdao', 'juliapkg', 'juliacall'],
   'keywords': ['openmdao_component'],
   'license': 'MIT',
   'name': 'omjlcomps',
   'packages': ['omjlcomps', 'omjlcomps.test'],
   'version': '0.1.5',
   'include_package_data': True}

setup(**setup_args)
